/*
This file is part of the IODEV (Input Output DEVices) component of BOCHS.

The IODEV component is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

The IODEV component is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the IODEV component; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/



#include <stdio.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "iodev.h"



/* For the 64K bytes of IO memory, no memory is actually allocated.
 * Instead, each time a read/write operation to IO memory occurs,
 * a device handler associated with the IO address is called to
 * perform the operation.  For those addresses which are not handled
 * by a device handler, a null handler is called and does nothing.
 *
 * In order to efficiently map an IO address to an IO handler,
 * I use 64K bytes of handler id's, each byte designating an id
 * (handle) for the device.  This id can be used as an index into
 * a table which includes the function pointer of the handler.
 */

Bit8u bx_io_handler_id[0x10000];  /* 64K */

struct {
  Bit8u (* funct) (Bit8u, Bit8u, Bit32u, Bit8u);    /* function pointer */
  char *        handler_name;                       /* name of device */
  int           irq;                                /* associated irq if any */
  Bit8u         address_space;                      /* IO or memory address */
  } bx_io_handler[BX_MAX_IO_DEVICES];

/* more for informative purposes, the names of the divices which
 * are use each of the IRQ 0..15 lines are stored here
 */
char *bx_irq_handler_name[BX_MAX_IRQS];

/* number of currently registered io handlers (devices) */
Bit8u bx_num_io_handles = 0;




/* for efficient repetitive checking of multiple ansynchronous
 * events, the bx_async_event flag is set if any async event occur.
 * That is, for timer events, file descriptor events, and by the PIC.
 */
Boolean bx_async_event = 0;
   
/* after an asynchronous event is determined, the following flags
 * determine which type of event(s) occured */
Boolean bx_timer_event = 0; /* the timer expired */
Boolean bx_fd_event    = 0; /* a file descriptor event occured */


/* define a variable signifying an event on the corresponding pins.
   INTR is the maskable interrupt pin.  on the x86, it is checked
     after completing every instruction to determine whether a hardware
     interrupt request is pending.  this check can be masked (blocked)
     by clearing the interrupt flag.
     The bx_INTR flag is set by the PIC.
     The bx_INTR_vector variable is also set by the PIC so the CPU can
     determine which interrupt occurred.
   NMI is he non-maskable interrupt pin.  Upon an NMI, the x86 aborts
     program execution immediately after completing the current instruction.
 */
Boolean bx_INTR = 0;
Bit8u   bx_INTR_vector;
Boolean bx_NMI  = 0;



/* for now, if an INTR occurs and is not masked, a C function is called
 * to handle the interrupt.  Functions register themselves with and
 * associated interrupt number 0..255 by means of bx_register_int_vector()
 * NOTE: IRQ's 0..7  = interrupt 08h..0Fh
 *       IRQ's 8..15 = interrupt 70h..77h
 */
void (*bx_interrupt_table[256])(int);


/* To allow IO devices to register functions to be called when events
 * occur on associated file descriptors, we need to define a set of
 * file descriptors to select() on, and a couple arrays to hold the
 * file descriptors, and the function pointers
 */
static fd_set fd_read_set, fd_write_set, fd_exception_set; /* for select() */
static int max_fd = 0; /* maximum fd (to pass to select() ) */

#define BX_MAX_REGISTERED_FDS  20

int registered_fds[BX_MAX_REGISTERED_FDS]; /* values of fd's registered */
void (* registered_fd_handler[BX_MAX_REGISTERED_FDS])(void); /* function ptrs*/
int num_registered_fds = 0; /* number of currently registered fd's */


/* local prototypes */
void bx_iodev_init(void);




  int
main(int argc, char *argv[])
{
  bx_iodev_init();

  bx_register_io_handlers();

  bx_bios_post();

  bx_cpu_loop();
}



/* The main CPU instruction execution loop */

  void
bx_cpu_loop(void)
{
  sigset_t set, oldset;
  unsigned i, b = 1;
  Bit8u val;
  int instr_cnt;
  fd_set read_set, write_set, exception_set;
  struct timeval timeout;

  sigemptyset( &set );
  sigaddset( &set, SIGVTALRM );


  val = bx_inp(0x21);
  bx_outp(0x21, val & ~(1<<0));



  instr_cnt = BX_CYCLES_PER_FD_CHECK;

  while (1) {

    /* since not all systems allow for asynchronous notification of
       file descriptor events (readable, writeable, exceptions, etc),
       and even for those that do, not all types of file descriptors
       are supported,we need to periodically induce
       a select on file descriptors registered by the bx_add_input()
       function.  for now, I'll just use a counter which is decremented
       before each instruction.  in the future, I might switch to using
       a timer to notify me periodically, then this extra check won't
       be necessary */
    instr_cnt--;
    if (!instr_cnt) {
      instr_cnt = BX_CYCLES_PER_FD_CHECK;
      bx_fd_event = 1;
      bx_async_event = 1;
      }

    if (bx_async_event) {
      /* block signals */
      sigprocmask( SIG_BLOCK, &set, &oldset);
      
      while (1) {
	if (bx_timer_event) {
	  bx_timer_event = 0;
	  bx_pit_handler(BX_CALL_REASON_TIMER, 0, 0, 0);
	  }
        else if (bx_fd_event) {
          bx_fd_event = 0;
          timeout.tv_sec = 0;
          timeout.tv_usec = 0;
          read_set = fd_read_set;
          write_set = fd_write_set;
          exception_set = fd_exception_set;
          if ( select(max_fd, &read_set, &write_set, &exception_set,
              &timeout) ) {
            for (i=0; i < num_registered_fds; i++) {
              if (FD_ISSET(registered_fds[i], &fd_read_set) ||
                  FD_ISSET(registered_fds[i], &fd_write_set) ||
                  FD_ISSET(registered_fds[i], &fd_exception_set) ) {
                /* call registered function for fd here */
                }
              }
            }
          }
        else {
          break; /* no more async events to process */
          }

	if (bx_INTR) {
          if (bx_interrupt_table[bx_INTR_vector]) {
            (*bx_interrupt_table[bx_INTR_vector])(bx_INTR_vector);
	    }
          bx_INTR = 0;
	  }
        else {
          }
        }
      
      bx_async_event = 0;
      /* unblock signals */
      sigprocmask( SIG_SETMASK, &oldset, (sigset_t *) NULL);
      }

    /* instruction emulation here */
    b = b << 1;  /* bogus operation */
    }
}


  int
bx_register_io_handler(bx_iodev_t device)
{
  int handle, i;

  if (device.irq != BX_NO_IRQ  &&  device.irq >= BX_MAX_IRQS) {
    bx_error(1, "IO device %s registered with IRQ=%d above %lu\n",
      device.handler_name, device.irq, (UL) BX_MAX_IRQS-1);
    }

  if (device.addr_space == BX_MEMORY_ADDRESS_SPACE) {
    /* currently only VGA can memory map IO.  see bx_peek(), bx_poke() */
    return(0);
    }

  /* first find existing handle for function or create new one */
  for (handle=0; handle < bx_num_io_handles; handle++) {
    if (bx_io_handler[handle].funct == device.io_handler) break;
    }
  if (handle >= bx_num_io_handles) {
    /* no existing handle found, create new one */
    if (bx_num_io_handles >= BX_MAX_IO_DEVICES) {
      bx_error(0, "too many IO devices installed.\n");
      bx_error(1, "  try increasing BX_MAX_IO_DEVICES\n");
      }
    if (device.irq != BX_NO_IRQ  &&  bx_irq_handler_name[device.irq]) {
      bx_error(1, "IRQ %d conflict.  IO devices %s & %s\n",
        device.irq, bx_irq_handler_name[device.irq], device.handler_name);
      }
    if (device.irq != BX_NO_IRQ)
      bx_irq_handler_name[device.irq] = device.handler_name;
    bx_num_io_handles++;
    bx_io_handler[handle].funct          = device.io_handler;
    bx_io_handler[handle].handler_name   = device.handler_name;
    bx_io_handler[handle].irq            = device.irq;
    bx_io_handler[handle].address_space  = device.addr_space;
    }

  /* change table to reflect new handler id for that address */
  for (i=device.start_addr; i <= device.end_addr; i++) {
    if (bx_io_handler_id[i] != 0) {
      bx_error(0, "IO device address conflict at IO address %lXh\n", (UL) i);
      bx_error(1, "  conflicting devices: %s & %s\n",
        bx_io_handler[handle].handler_name, bx_io_handler[bx_io_handler_id[i]].handler_name);
      }
    bx_io_handler_id[i] = handle;
    }
}



/* The master event (signal) handler function. Currently this only
   handles the timer signal
 */

  void
bx_master_event_handler(int sig)
{
  /* only reinstate the signal handler for those platforms that can't be
     told to automatically */
#ifndef HAVE_SIGACTION
  signal(sig, bx_master_event_handler);
#endif


  if (sig == SIGVTALRM) {
    bx_timer_event = 1;
    bx_async_event = 1;
    }
}



/* bx_error()
 *
 * To print warnings, errors, etc, use the variable argument list
 * bx_error() function.
 *
 * bomb_out       0 = continue, 1 = exit
 * ...            the format and parameters typically passed to printf()
 */

  void
bx_error(int bomb_out, ...)
{
  va_list ap;
  char *fmt;

  fprintf(stderr, "bochs: error, ");
  va_start(ap, bomb_out);
  fmt = va_arg(ap, char *);
  vfprintf(stderr, fmt, ap);
  va_end(ap);

  if (bomb_out != 0) exit(bomb_out);
}


/* bx_outp()
 *
 * Write a byte of data to the IO memory address space.
 */

  void
bx_outp(Bit16u addr, Bit8u value)
{
  Bit8u handle;

  handle = bx_io_handler_id[addr];
  (* bx_io_handler[handle].funct)(BX_CALL_REASON_WRITE,
    BX_IO_ADDRESS_SPACE, (Bit32u) addr, value);
}



/* bx_inp()
 *
 * Read a byte of data from the IO memory address space
 */

  Bit8u
bx_inp(Bit16u addr)
{
  Bit8u handle;
  Bit8u ret;

  handle = bx_io_handler_id[addr];
  ret = (* bx_io_handler[handle].funct)(BX_CALL_REASON_READ,
    BX_IO_ADDRESS_SPACE, (Bit32u) addr, 0);
  return(ret);
}


/* bx_peek()
 *
 * read a byte of data from main memory address space
 */

  Bit8u
bx_peek(Bit32u addr)
{
  Bit8u ret;

  if (addr < 0xA0000  ||  addr > 0xBFFFF) {
    bx_error(0, "bx_peek(): read outside defined IO memory space\n");
    return;
    }
  ret = bx_vga_handler(BX_CALL_REASON_READ,
    BX_MEMORY_ADDRESS_SPACE, addr, 0);
  return(ret);
}


/* bx_poke()
 *
 * write a byte of data to main memory address space
 */

  void
bx_poke(Bit32u addr, Bit8u value)
{
  if (addr < 0xA0000  ||  addr > 0xBFFFF) {
    bx_error(0, "bx_poke(): write outside defined IO memory space\n");
    return;
    }
  bx_vga_handler(BX_CALL_REASON_WRITE,
    BX_MEMORY_ADDRESS_SPACE, addr, value);
}



/* bx_register_int_vector()
 *
 * register a C function to be called when there is an interrupt
 * designated by vector.  The bios_addr is currently unused, but
 * will in the future designate a ROM based address. more later...
 */

  void
bx_register_int_vector(Bit8u vector, Bit32u bios_addr,
    void (* funct)(int vector))
{
  /* ignoring bios_addr for now */

  bx_interrupt_table[vector] = funct;
}


 
/* bx_iodev_init()
 *
 * intialize local constructs used to all the IO device handling features
 */

 void
bx_iodev_init(void)
{
  unsigned i;

  /* handle 0 maps to the unmapped IO device handler.  Basically any
     ports which don't map to any other device get mapped to this
     handler which does absolutely nothing.
   */
  bx_io_handler[0].funct          = bx_unmapped_io_handler;
  bx_io_handler[0].handler_name   = "unmapped io";
  bx_io_handler[0].irq            = BX_NO_IRQ;
  bx_io_handler[0].address_space  = BX_IO_ADDRESS_SPACE;
  bx_num_io_handles = 1;

  /* set unused elements to appropriate values */
  for (i=1; i < BX_MAX_IO_DEVICES; i++) {
    bx_io_handler[i].funct          = NULL;
    }

  for (i=0; i < 0x10000; i++)
    bx_io_handler_id[i] = 0;  /* unmapped IO handle */

  for (i=0; i < BX_MAX_IRQS; i++) {
    bx_irq_handler_name[i] = NULL;
    }

  for (i=0; i < 256; i++) {
    bx_interrupt_table[i] = NULL;
    }

  FD_ZERO(&fd_read_set);
  FD_ZERO(&fd_write_set);
  FD_ZERO(&fd_exception_set);
}



/* bx_add_input()
 *
 * register a C function to be called when there is an event on the
 * associated file descriptor fd.
 * 
 * event_type is comprised of the bit-wise or of one or more of the
 * following flags:
 *   BX_FD_EVENT_READ:          per select(), a read event
 *   BX_FD_EVENT_WRITE:         per select(), a write event
 *   BX_FD_EVENT_EXCEPTION:     per select(), an exception event
 */

  void
bx_add_input(int fd, int event_type, void (* funct)(void))
{
  if (num_registered_fds >= BX_MAX_REGISTERED_FDS) return;

  registered_fds[num_registered_fds] = fd;
  registered_fd_handler[num_registered_fds] = funct;
  num_registered_fds++;

  if (event_type & BX_FD_EVENT_READ) {
    FD_SET(fd, &fd_read_set);
    }
  if (event_type & BX_FD_EVENT_WRITE) {
    FD_SET(fd, &fd_write_set);
    }
  if (event_type & BX_FD_EVENT_EXCEPTION) {
    FD_SET(fd, &fd_exception_set);
    }

  if (fd >= max_fd)
    max_fd = fd + 1;
}
